grammar TMI;

@header
{
package com.dexvis.tmi;
}

program
	: statement* EOF;

statement
	: foreachLoop
	| blockStatement
	| ifStatement
	| forLoop
	| whileLoop
	| expression (('if'|'unless') expression)? ';';

blockStatement
	: '{' statement* '}';

expressionStatement
	: ';'
	| expression ';';

forLoop
	: 'for' '(' expressionStatement expressionStatement expression ')' statement;

whileLoop
	: 'while' '(' expression ')' statement;

ifStatement
	: 'if' '(' expression ')' statement ('else' statement)?;

foreachLoop
	: 'foreach' IDENTIFIER '(' range ')' statement;

range
	: integerLiteral ('..' integerLiteral)?		# intRange
	| charLiteral ('..' charLiteral)?			# charRange
	| stringLiteral ('..' stringLiteral)?		# stringRange
	| range (',' range)+						# rangeUnion;

expression
	: assignmentExpression;

// Conflicts
//	| expression ',' assignmentExpression;

assignmentExpression
	: conditionalExpression
	| assignmentExpression assignmentOperator assignmentExpression;

assignmentOperator
	: '=' | '*=' | '/=' | '%=' | '+=' | '-=' | '<<=' | '>>=' | '&=' | '^=' | '|=';

conditionalExpression
	: logicalExpression
	| conditionalExpression '?' conditionalExpression ':' conditionalExpression;

logicalExpression
	: bitwiseExpression
	| logicalExpression ('&&'|'||') logicalExpression;

bitwiseExpression
	: relationalExpression
	| bitwiseExpression ('^'|'|'|'&') bitwiseExpression;

relationalExpression
	: shiftExpression
	| relationalExpression ('=='|'!='|'<='|'>='|'<'|'>'|'=~') relationalExpression;

shiftExpression
	: additiveExpression
	| shiftExpression ('<<'|'>>') shiftExpression;

additiveExpression
	: multiplicativeExpression
	| additiveExpression ('+'|'-') additiveExpression;

multiplicativeExpression
	: unaryExpression
	| multiplicativeExpression ('*'|'/'|'%') multiplicativeExpression;

unaryExpression
	: postfixExpression
	| ('++'|'--'|'+'|'-') unaryExpression;

postfixExpression
	: tableReference
	| identifier methodParameters
	| postfixExpression '++'
	| postfixExpression '--';

tableReference
	: primaryExpression
	| tableReference '[' (range|columnAliasList) ']' ('[' range ']')?;

primaryExpression
	:  identifier
	| dateLiteral
	| floatLiteral
	| integerLiteral
	| stringLiteral
	| booleanLiteral
	| charLiteral
	| regexLiteral
	| tableLiteral
	| mapLiteral
	| '(' expression ')'
	| method
	| builtin;

method
	: '{' methodVariables statement* '}';

methodVariables
	: (identifier (',' identifier)* '->')?;

methodParameters
	: '(' (primaryExpression (',' primaryExpression)*)? ')';

identifier
	: (columnIdentifier|IDENTIFIER) ('as' ('Integer'|'Boolean'|'Float'|'String'|'Character'))?;

columnIdentifier
	: IDENTIFIER '.' columnAlias;

floatLiteral : FLOAT_LITERAL;

integerLiteral : INTEGER_LITERAL;

stringLiteral : STRING_LITERAL;

booleanLiteral : BOOLEAN_LITERAL;

charLiteral : CHARACTER_LITERAL;

regexLiteral : REGEX_LITERAL;

dateLiteral
	: 'date' '(' (expression (',' expression)*)? ')';

builtin
	: ('print'|'println') expression 	#print
	| 'split' columnAliasList
		'with' regexLiteral
		'from' identifier				#splitColumns
	| 'select' columnAliasList 'from' identifier
	  whereClause? whenClause*			#selectColumns
	| 'join' (columnAliasList 'from')? identifier ',' identifier
	  ( 'where' expression )?			#join
	| 'return' '(' expression ')'		#returnValue
	| ('cos'|'sin'|'tan'|'acos'|'asin'|'atan'|
	   'cosh'|'sinh'|'tanh'|'abs'|'ceiling'|
	   'floor'|'log'|'log10'|'random'|'round'|
	   'sqrt'|'toRadians'|'toDegrees')
	   ('(' expression ')' | expression)	#math1;

whereClause
	: 'where' expression;

whenClause
	: 'when' expression '{' statement* '}';

mapLiteral
	: '[' mapNvps ']';

mapNvps
	: (mapNvp (',' mapNvp)*)?;

mapNvp
	: expression ':' expression;

tableLiteral
	: '[' header ',' tableData ']';

header
	: '[' (columnLabel (',' columnLabel)*)? ']';

columnAliasList
	: columnAlias (',' columnAlias)*;

columnAlias
	: columnLabel ('rename' columnLabel)?;

columnLabel
	: identifier | stringLiteral;

tableData
	: '[' (tableRow (',' tableRow)*)? ']';

tableRow
	: '[' (tableCell (',' tableCell)*)? ']';

tableCell
	: identifier | integerLiteral | stringLiteral | booleanLiteral | floatLiteral | charLiteral;

//////////
// Tokens
//////////

SINGLE_LINE_COMMENT :   '#' ~('\n'|'\r')* '\r'? '\n' {skip();};

fragment UPPER
	: 'A'..'Z'|'$'|'_';

fragment LOWER
	: 'a'..'z';

fragment LETTER
	: UPPER | LOWER;
	
fragment DIGIT
	: NONZERO_DIGIT | '0';

fragment NONZERO_DIGIT
	: '1'..'9';

fragment OCTAL_DIGIT
	: '0'..'7';

INTEGER_LITERAL
	: (DECIMAL_NUMERAL|HEX_NUMERAL|OCTAL_NUMERAL) ('L'|'l')?;

DECIMAL_NUMERAL
	: DIGIT+;

HEX_NUMERAL
	: '0' 'x' HEX_DIGIT+;

fragment HEX_DIGIT
	: DIGIT | 'A'..'F' | 'a'..'f';

OCTAL_NUMERAL
	: '0' OCTAL_DIGIT+;

FLOAT_LITERAL	: DIGIT+ ('.' DIGIT+)? EXPONENT_PART? FLOAT_TYPE?
				| '.' DIGIT+ EXPONENT_PART? FLOAT_TYPE?
				| DIGIT+ EXPONENT_PART FLOAT_TYPE?
				| DIGIT+ EXPONENT_PART? FLOAT_TYPE;

EXPONENT_PART	: ('E'|'e') ('+'|'-')? DIGIT+;

FLOAT_TYPE		: 'F' | 'f' | 'D' | 'd';

WS  : [ \t\n\r\f]+ -> skip ;

BOOLEAN_LITERAL
	: 'true' | 'false';

IDENTIFIER
    :    (LETTER | '_' | '$') (LETTER | DIGIT | '_')*;

CHARACTER_LITERAL
	:   '\'' ( ESCAPE_SEQUENCE | ~('\''|'\\') ) '\'';

STRING_LITERAL
	:  '"' ( ESCAPE_SEQUENCE | ~('\\'|'"') )* '"';

REGEX_LITERAL
	:  '/' ( REGEX_ESCAPE_SEQUENCE | ~('\\'|'/') )* '/';

fragment REGEX_ESCAPE_SEQUENCE
    :   '\\' ('t'|'n'|'r'|'f'|'a'|'e'|'c'|'d'|'D'|'s'|'S'|'w'|'W'|'p'|'P'|'b'|'B'|'A'|'G'|'z'|'Z'|'Q'|'E'|'\"'|'\''|'\\')
    |   OCTAL_ESCAPE
    ;

fragment ESCAPE_SEQUENCE
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   OCTAL_ESCAPE
    ;

fragment OCTAL_ESCAPE
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;
